#!/usr/bin/env python3
# -*- coding: utf-8 -*-

"""
Functions to process inputs for Neural Network model

@author: vincent
"""

import os
import sys
import copy
import csv
import numpy as np
import time
import netCDF4 as nc
from scipy.stats import qmc
from scipy.interpolate import griddata,interp1d

def pastPeriodMean(timing,period,values):
    '''
    Computes the mean of values over past period at each step of timing
    '''
    ntot         = len(timing)
    deltat       = np.mean(np.diff(timing))
    if(np.any(abs(np.diff(timing)[1:-1]-deltat)>deltat/10)):
        print('Error: pastPeriodMean() only implemented for regularly-spaced time series')
        sys.exit()
    ninprd       = int(np.floor(period/deltat))
    valsinperiod = np.zeros_like(values)
    if(len(np.shape(values))==1):
        for tt in range(ninprd-1,ntot):
            valsinperiod[tt] = np.mean(values[tt-(ninprd-1):tt+1])
    elif(len(np.shape(values))==2):
        if(np.shape(values)[0]==ntot):
            for ii in range(np.shape(values)[1]):
                for tt in range(ninprd-1,ntot):
                    valsinperiod[tt,ii] = np.mean(values[tt-(ninprd-1):tt+1,ii])
        elif(np.shape(values)[1]==ntot):
            for ii in range(np.shape(values)[0]):
                for tt in range(ninprd-1,ntot):
                    valsinperiod[ii,tt] = np.mean(values[ii,tt-(ninprd-1):tt+1])
    return(valsinperiod)

def pastPeriodMean3dTensor(timing,period,values,axistime):
    '''
    Computes the mean of values over past period at each step of timing
    '''
    ntot         = len(timing)
    deltat       = np.mean(np.diff(timing))
    if(np.any(abs(np.diff(timing)[1:-1]-deltat)>deltat/10)):
        print('Error: pastPeriodMean() only implemented for regularly-spaced time series')
        sys.exit()
    ninprd       = int(np.floor(period/deltat))
    values       = np.moveaxis(values,axistime,0) #put time axis on axis 0
    valsinperiod = np.zeros_like(values)
    for tt in range(ninprd-1,ntot):
        valsinperiod[tt,:,:] = np.mean(values[tt-(ninprd-1):tt+1,:,:],axis=0)
    valsinperiod = np.moveaxis(valsinperiod,0,axistime) #reset original axes
    return(valsinperiod)

def pastBoundedPeriodMean3dTensor(timing,pastStart,pastEnd,values,axistime):
    '''
    Computes the mean of values over past period at each step of timing
    The past period is bounded within [pastStart:pastEnd]
    Note that pastStart and pastEnd should be negative values
    The value of pastStart should be more negative than pastEnd
    '''
    ntot         = len(timing)
    if(pastStart>0 or pastEnd>0):
        ValueError('pastStart and pastEnd should be negative in pastBoundedPeriodMean3dTensor()')
    if(pastStart>=pastEnd):
        ValueError('pastStart must be more negative than pastEnd in pastBoundedPeriodMean3dTensor()')
    values       = np.moveaxis(values,axistime,0) #put time axis on axis 0
    valsinperiod = np.zeros_like(values)
    for tt in range(ntot):
        inds = np.logical_and(timing>=timing[tt]+pastStart,timing<timing[tt]+pastEnd)
        if(np.sum(inds)>=1):
            valsinperiod[tt,:,:] = np.mean(values[inds,:,:],axis=0)
    valsinperiod = np.moveaxis(valsinperiod,0,axistime) #reset original axes
    return(valsinperiod)

def pastPeriodTotal(timing,period,values):
    '''
    Counts the total of values over past period at each step of timing
    '''
    ntot         = len(timing)
    deltat       = np.mean(np.diff(timing))
    ninprd       = int(np.floor(period/deltat))
    valsinperiod = np.zeros_like(values)
    if(len(np.shape(values))==1):
        for tt in range(ninprd,ntot):
            valsinperiod[tt] = np.sum(values[tt-ninprd:tt])
    elif(len(np.shape(values))==2):
        if(np.shape(values)[0]==ntot):
            for ii in range(np.shape(values)[1]):
                for tt in range(ninprd,ntot):
                    valsinperiod[tt,ii] = np.sum(values[tt-ninprd:tt,ii])
        elif(np.shape(values)[1]==ntot):
            for ii in range(np.shape(values)[0]):
                for tt in range(ninprd,ntot):
                    valsinperiod[ii,tt] = np.sum(values[ii,tt-ninprd:tt])
    return(valsinperiod)

def countInRadius(xgrid,ygrid,binarypresence,radius):
    '''
    Gets the total count of occurences of binarypresence within radius
    Calculated at each gridpoint of (xgrid,ygrid)
    '''
    ntot         = len(xgrid)
    ncounts      = np.zeros(ntot)
    for ii in range(ntot):
        dists       = np.sqrt((xgrid[ii]-xgrid)**2+(ygrid[ii]-ygrid)**2)
        inds        = dists<=radius
        ncounts[ii] = np.sum(binarypresence[inds])
    return(ncounts)

def sdevInRadius(xgrid,ygrid,values,radius):
    '''
    Gets the standard deviation of values within radius
    Calculated at each gridpoint of (xgrid,ygrid)
    '''
    ntot         = len(xgrid)
    fullsdevs    = np.zeros(ntot)
    for ii in range(ntot):
        dists         = np.sqrt((xgrid[ii]-xgrid)**2+(ygrid[ii]-ygrid)**2)
        inds          = dists<=radius
        fullsdevs[ii] = np.std(values[inds])
    return(fullsdevs)

def minDistToPoints(xgrid,ygrid,binarypresence):
    '''
    Gets the minimal distance to a positive binarypresence
    Calculated at each gridpoint of (xgrid,ygrid)
    '''
    ntot         = len(xgrid)
    distmin      = np.zeros(ntot)
    for ii in range(ntot):
        dists                    = np.sqrt((xgrid[ii]-xgrid)**2+(ygrid[ii]-ygrid)**2)
        dists[binarypresence==0] = max(dists)
        distmin[ii]              = min(dists)
    return(distmin)

def selectPointsHomogeneously(xgrid,ygrid,numpoints,ninxdir,ninydir,borderbuffer=0):
    '''
    Selects numpoints points homogeneously within the (xgrid,ygrid) grid
    Number of points along the x- and y-directions should be given
    A buffer zone away from domain boundaries can be provided (default:0)
    '''
    if(ninxdir*ninydir!=numpoints):
        print('Error: ninxdir*ninydir!=numpoints in selectPointsHomogeneously()')
        sys.exit()
    posx    = np.concatenate([np.linspace(min(xgrid)+borderbuffer,max(xgrid)-borderbuffer,ninxdir) for kk in range(ninydir)])
    posy    = np.concatenate([np.linspace(min(ygrid)+borderbuffer,max(ygrid)-borderbuffer,ninydir) for kk in range(ninxdir)])
    nsel    = len(posx)
    inds    = np.zeros(nsel).astype(int)
    for kk in range(nsel):
        inds[kk] = np.argmin(np.sqrt((xgrid-posx[kk])**2+(ygrid-posy[kk])**2))
    return(inds)

def latinhypercubesampler(numpoints,lsArrays,getIndices=False):
    '''
    Latin hypercube of numpoints points
    Dimensionality of the space is given by number of arrays passed throug lsArrays
    Lower and upper bounds are taken from each individual array
    Gets values within [lowerbound[dd],upperbound[dd]] for each dimension dd
    Returns values of the selection
    Optionally also returns indices of the selection (but all arrays must be of same size)
    '''
    dim     = len(lsArrays)
    sampler = qmc.LatinHypercube(d=dim)
    sample  = sampler.random(n=numpoints)
    lowerbounds = np.array([np.min(lsArrays[dd]) for dd in range(0,dim)])
    upperbounds = np.array([np.max(lsArrays[dd]) for dd in range(0,dim)])
    vals        = qmc.scale(sample,lowerbounds,upperbounds)
    if(min([len(lsArrays[dd]) for dd in range(0,dim)])<numpoints):
        print('Error: numpoints larger than some arrays in latinhypercubesampler')
        sys.exit()
    if(getIndices==False):
        return(vals)
    else:
        szarr = np.array([len(lsArrays[dd]) for dd in range(0,dim)])
        if(np.any(szarr!=szarr[0])):
            print('Error: indices are requested in latinhypercubesampler but arrays have different sizes')
            sys.exit()
        inds = np.array([]).astype(int)
        # Standardize each array for distance computations #
        mnArrays   = np.array([np.mean(lsArrays[dd]) for dd in range(0,dim)])
        sdevArrays = np.array([np.std(lsArrays[dd]) for dd in range(0,dim)])
        lsStd      = [(lsArrays[dd]-mnArrays[dd])/sdevArrays[dd] for dd in range(0,dim)] #standardized arrays
        matmn      = np.vstack([np.repeat(mnArrays[dd],numpoints) for dd in range(0,dim)]).T #concatenate arrays in matrix column-wise
        matsdev    = np.vstack([np.repeat(sdevArrays[dd],numpoints) for dd in range(0,dim)]).T #concatenate arrays in matrix column-wise
        valsStd    = (vals-matmn)/matsdev #values standardized
        matArrStd  = np.vstack(lsStd).T #concatenate arrays in matrix column-wise
        for rr in range(numpoints):
            dists       = np.sqrt(np.sum((matArrStd-valsStd[rr,:])**2,axis=1))
            dists[inds] = 1e20 #rejecting points already selected
            inds        = np.append(inds,np.argmin(dists))
        return(vals,inds)
            

def arrayStandardize(values,getMeanAndSdev=False):
    '''
    Standardization
    '''
    inonan = np.isnan(values)==False
    mnval = np.mean(values[inonan])
    sdev  = np.std(values[inonan])
    output = (values-mnval)/sdev
    if(getMeanAndSdev):
        return(output,mnval,sdev)
    else:
        return(output)

def arrayLogStandardize(values,epsilon,getMeanAndSdev=False):
    '''
    Log-Standardization
    '''
    inonan = np.isnan(values)==False
    mnval = np.mean(values[inonan])
    sdev  = np.std(values[inonan])
    output = np.log(epsilon+values)-np.log(epsilon) #see Rasper et al. (2021)
    if(getMeanAndSdev):
        return(output,mnval,sdev)
    else:
        return(output)

def arrayNormalize(values,minval=-1,maxval=1):
    '''
    Normalization of values between chosen bounds
    '''
    vmax = np.max(values)
    vmin = np.min(values)
    if(minval==-1 and maxval==1):
        nrmvalues = -1+2*(values-vmin)/(vmax-vmin)
    elif(minval==0 and maxval==1):
        nrmvalues = (values-vmin)/(vmax-vmin)
    else:
        print(f'minval={minval},maxval={maxval} not supported in arrayNormalize()')
    return(nrmvalues)

def regrid2Dnonan(nnx,nny,xx1d,yy1d,vals1d,mostsquashed='x',methodinterp='linear',caution=None):
    '''
    From 1-dimensional x and y vectors (xx1d,yy1d)
    Creates a 2-dimensional grid of size nny*nnx
    Grid is adjusted such that no values interpolated from vals1d on the grid is nan
    Returns xcoordsongrid,ycoordsongrid
    '''
    ### Initial regular 2D grid ###
    xlvmin    = 1e3*(np.floor(min(xx1d)*1e-3))
    xlvmax    = 1e3*(np.ceil(max(xx1d)*1e-3))
    ylvmin    = 1e3*(np.floor(min(yy1d)*1e-3))
    ylvmax    = 1e3*(np.ceil(max(yy1d)*1e-3))
    xlvs      = np.linspace(xlvmin,xlvmax+1,nnx)
    ylvs      = np.linspace(ylvmin,ylvmax+1,nny)
    gridx0   = np.zeros((nny,nnx));
    gridy0   = np.zeros((nny,nnx));
    for ii in range(nny):
        gridx0[ii,:] = xlvs;
    for jj in range(nnx):
        gridy0[:,jj] = ylvs;
            
    ### Interpolate ###
    gridvals0 = griddata(np.vstack((yy1d,xx1d)).T,vals1d,(gridy0,gridx0),methodinterp)

    ### Keep only rows with >1 non-nan #
    gridx1   = gridx0[[np.sum(np.isnan(gridvals0[rr,:])==False)>1 for rr in range(nny)],:]
    gridy1   = gridy0[[np.sum(np.isnan(gridvals0[rr,:])==False)>1 for rr in range(nny)],:]
    gridvals1 = griddata(np.vstack((yy1d,xx1d)).T,vals1d,(gridy1,gridx1),methodinterp)
    ### Keep only cols with >1 non-nan ###
    gridx1   = gridx1[:,[np.sum(np.isnan(gridvals1[:,cc])==False)>1 for cc in range(nnx)]]
    gridy1   = gridy1[:,[np.sum(np.isnan(gridvals1[:,cc])==False)>1 for cc in range(nnx)]]
      
    ### Regrid ###
    xlvmin   = np.min(gridx1)
    xlvmax   = np.max(gridx1)
    ylvmin   = np.min(gridy1)
    ylvmax   = np.max(gridy1)
    xlvs     = np.linspace(xlvmin,xlvmax+1,nnx)
    ylvs     = np.linspace(ylvmin,ylvmax+1,nny)
    gridx1   = np.zeros((nny,nnx))
    gridy1   = np.zeros((nny,nnx))
    for ii in range(nny):
        gridx1[ii,:] = xlvs
    for jj in range(nnx):
        gridy1[:,jj] = ylvs
    gridvals1 = griddata(np.vstack((yy1d,xx1d)).T,vals1d,(gridy1,gridx1),methodinterp)
    gridx2 = np.copy(gridx1)
    gridy2 = np.copy(gridy1)
    
    if(mostsquashed=='x'):
        # Squash each row between non-nan values #
        for rr in range(nny):
            if(np.any(np.isnan(gridvals1[rr,:]))):
                cc0 = np.where(np.isnan(gridvals1[rr,:])==False)[0][0]
                cc1 = np.where(np.isnan(gridvals1[rr,:])==False)[0][-1]
                rrx = np.linspace(gridx1[rr,cc0],gridx1[rr,cc1],nnx)
                gridx2[rr,:] = np.copy(rrx)
    elif(mostsquashed=='y'):
        # Squash each column between non-nan values #
        for cc in range(nnx):
            if(np.any(np.isnan(gridvals1[:,cc]))):
                rr0 = np.where(np.isnan(gridvals1[:,cc])==False)[0][0]
                rr1 = np.where(np.isnan(gridvals1[:,cc])==False)[0][-1]
                ccy = np.linspace(gridy1[rr0,cc],gridy1[rr1,cc],nny)
                gridy2[:,cc] = np.copy(ccy)
    else:
        print("Error: regrid2Dnonan() only accepts 'x' or 'y' as mostsquashed argument")
    
    gridvals2 = griddata(np.vstack((yy1d,xx1d)).T,vals1d,(gridy2,gridx2),methodinterp)
    if(np.any(np.isnan(gridvals2))):
        print('regrid2Dnonan failed')
    return(gridx2,gridy2)

def samplingInTime1dArray(timing,samplingperiod,values,method):
    '''
    Computes the mean of values over past period at each step of timing
    '''
    
    PossibleMethods = ['interpolation','averaging','cumulative']
    if(method not in PossibleMethods):
        ValueError(f'Method {method} not implemented in samplingInTime1dArray')

    ntot         = len(timing)
    if(len(values)!=ntot):
        ValueError('Number of values and lenght of timing do not agree in samplingInTime1dArray')
        
    deltat       = np.mean(np.diff(timing))
    if(np.any(abs(np.diff(timing)[1:-1]-deltat)>deltat/10)):
        ValueError('samplingInTime1dArray() only implemented for regularly-spaced time series')

    ninprd        = int(np.floor(samplingperiod/deltat))
    
    ### Interpolating at midpoint of each sampling period ###
    if(method=='interpolation'):
        steps         = np.arange(int(ninprd/2),ntot,ninprd)
        tmnew         = timing[steps]
        ntotnew       = len(steps)
        lininterp     = interp1d(timing,values,kind='linear') #linear interpolator
        valsatnewtime = lininterp(steps)
    
    ### Taking mean over each sampling period ###
    if(method=='averaging'):
        steps         = np.arange(0,ntot,ninprd)
        tmnew         = timing[steps]+deltat/2
        ntotnew       = len(steps)
        valsatnewtime = np.zeros(ntotnew)
        for tt,stp0 in enumerate(steps):
            if(stp0!=steps[-1]):
                stp1 = steps[tt+1]
                valsatnewtime[tt] = np.mean(values[stp0:stp1])
            else:
                valsatnewtime[tt] = np.mean(values[stp0:])

    ### Taking total of values over each sampling period ###
    if(method=='cumulative'):
        steps         = np.arange(0,ntot,ninprd)
        tmnew         = np.append(timing[steps[1:]],timing[steps[-1]]+deltat)
        ntotnew       = len(steps)
        valsatnewtime = np.zeros(ntotnew)
        for tt,stp0 in enumerate(steps):
            if(stp0!=steps[-1]):
                stp1 = steps[tt+1]
                valsatnewtime[tt] = np.sum(values[stp0:stp1])
            else:
                valsatnewtime[tt] = np.sum(values[stp0:])

    return(tmnew,valsatnewtime)        
        
def samplingInTime2dArray(timing,samplingperiod,values,axistime,method):
    '''
    Computes the mean of values over past period at each step of timing
    '''
    
    PossibleMethods = ['interpolation','averaging','cumulative']
    if(method not in PossibleMethods):
        ValueError(f'Method {method} not implemented in samplingInTime2dArray')

    ntot         = len(timing)
    if(np.shape(values)[axistime]!=ntot):
        ValueError('Number of values and lenght of timing do not agree in samplingInTime2dArray')
        
    deltat       = np.mean(np.diff(timing))
    if(np.any(abs(np.diff(timing)[1:-1]-deltat)>deltat/10)):
        ValueError('samplingInTime2dArray() only implemented for regularly-spaced time series')

    ninprd        = int(np.floor(samplingperiod/deltat))
    values        = np.moveaxis(values,axistime,0) #put time axis on axis 0'
    otherdim      = np.shape(values)[1]
    
    ### Interpolating at midpoint of each sampling period ###
    if(method=='interpolation'):
        steps         = np.arange(int(ninprd/2),ntot,ninprd)
        tmnew         = timing[steps]
        ntotnew       = len(steps)
        valsatnewtime = np.zeros((ntotnew,otherdim))
        for kk in range(otherdim):
            lininterp           = interp1d(timing,values[:,kk],kind='linear') #linear interpolator
            valsatnewtime[:,kk] = lininterp(steps)
    
    ### Taking mean over each sampling period ###
    if(method=='averaging'):
        steps         = np.arange(0,ntot,ninprd)
        ntotnew       = len(steps)
        tmnew         = np.array([timing[0]+(ninprd/2)*deltat+kk*samplingperiod for kk in range(ntotnew)])
        valsatnewtime = np.zeros((ntotnew,otherdim))
        for tt,stp0 in enumerate(steps):
            if(stp0!=steps[-1]):
                stp1 = steps[tt+1]
                valsatnewtime[tt,:] = np.mean(values[stp0:stp1,:],axis=0)
            else:
                valsatnewtime[tt,:] = np.mean(values[stp0:,:],axis=0)

    ### Taking total of values over each sampling period ###
    if(method=='cumulative'):
        steps         = np.arange(0,ntot,ninprd)
        ntotnew       = len(steps)
        tmnew         = np.array([timing[0]+(ninprd/2)*deltat+kk*samplingperiod for kk in range(ntotnew)])
        valsatnewtime = np.zeros((ntotnew,otherdim))
        for tt,stp0 in enumerate(steps):
            if(stp0!=steps[-1]):
                stp1 = steps[tt+1]
                valsatnewtime[tt,:] = np.sum(values[stp0:stp1,:],axis=0)
            else:
                valsatnewtime[tt,:] = np.sum(values[stp0:,:],axis=0)

    valsatnewtime = np.moveaxis(valsatnewtime,0,axistime) #reset original axes
    return(tmnew,valsatnewtime)        
        

